﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.Build.Utilities;
using Microsoft.Build.Framework;
using System.Collections;
using System.Globalization;
using Microsoft.TeamFoundation;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace MsdnMag.Exchange.PushNotificationClient
{
    /// <summary>
    /// Based on Microsoft.TeamFoundation.Build.Tasks.CreateNewWorkItem.
    /// </summary>
    public class CreateNewWorkItem : Task
    {
        #region Constants

        private const string DEFAULT_WORKITEM_TYPE = "Bug";
        private const string DEFAULT_REASON = "New";

        #endregion

        #region Private Fields

        private string _description;
        private string _properties;
        private string _workItemTeamProject;
        private string _title;
        private string _workItemTypeName;
        private string _tfsUrl;
        private int _changesetId;
        private string _assignedTo;
        private int _workItemId;
        private int _numErrors;
        private TfsTeamProjectCollection _tfs;
        private VersionControlServer _versionControlServer;
        private WorkItemStore _workItemStore;
        private string _associatedWorkItems;
        private string[] _attachmentPaths;

        #endregion

        #region Methods

        public override bool Execute()
        {
            WorkItem workItem = new WorkItem(this.GetWorkItemType(this.WorkItemTeamProject, this.WorkItemType));
            workItem.Title = this.Title;
            workItem.History = this.Description;

            this.UpdateUserSpecifiedFields(workItem, PropertyParser.GetTable(this.WorkItemFieldValues));

            // Add a link to the changeset if one is specified:
            if (this.ChangesetId > 0)
            {
                AddLinkToChangeset(this.ChangesetId, workItem);
            }

            if (!string.IsNullOrWhiteSpace(this.AssociatedWorkItems))
            {
                this.AddLinkToAssociatedWorkItems(workItem, this.AssociatedWorkItems);
            }

            // Add any attachments specified:
            if (this.AttachmentPaths != null && this.AttachmentPaths.Length > 0)
            {
                this.AddAttachmentsToWorkItems(workItem, this.AttachmentPaths);
            }

            this.SaveWorkItem(workItem);

            Log.LogMessage("CreateNewWorkItem: Workitem Created: Id: {0}.", new object[] { workItem.Id });

            return true;
        }

        private void AddAttachmentsToWorkItems(WorkItem workItem, string[] attachmentPaths)
        {
            if (workItem == null || AttachmentPaths.Length == 0)
                return;

            foreach (var attachmentPath in attachmentPaths)
            {
                workItem.Attachments.Add(new Attachment(attachmentPath, ""));
            }
        }

        private void AddLinkToChangeset(int changesetId, WorkItem workItem)
        {
            Changeset changeset = this.VersionControlServer.GetChangeset(changesetId);
            Link link = new ExternalLink(this.WorkItemStore.RegisteredLinkTypes["Source Code File"], changeset.ArtifactUri.AbsoluteUri);
            workItem.Links.Add(link);
        }

        private void AddLinkToAssociatedWorkItems(WorkItem workItem, string associatedWorkItems)
        {
            string[] associatedWorkItemsArray = associatedWorkItems.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);

            if (associatedWorkItemsArray != null && associatedWorkItemsArray.Length > 0)
            {
                foreach (var item in associatedWorkItemsArray)
                {
                    try
                    {
                        WorkItem associatedWorkItem = this.WorkItemStore.GetWorkItem(Convert.ToInt32(item));
                        RelatedLink link = new RelatedLink(associatedWorkItem.Id);
                        workItem.Links.Add(link);

                        Log.LogMessage(string.Format("CreateNewWorkItem: Added a link to associated work item: {0}.", new object[] { associatedWorkItem.Id }));
                    }
                    catch (Exception ex)
                    {
                        Log.LogWarning(string.Format("Unable to associate work item: {0}.\r\nError Message: {1}\r\nStackTrace: {2}.", new object[] { item, ex.Message, ex.StackTrace }));
                    }
                }
            }
            else
            {
                Log.LogWarning("CreateNewWorkItem: AssociatedWorkItems list is empty.");
                return;
            }
        }

        private WorkItemType GetWorkItemType(string projectName, string type)
        {
            if (string.IsNullOrEmpty(type))
            {
                type = DEFAULT_WORKITEM_TYPE;
                Log.LogWarning(string.Format("CreateNewWorkItem: Using DefaultWorkItemType: {0}.", new object[] { type }));
            }

            try
            {
                Project project = this.WorkItemStore.Projects[projectName];
                return project.WorkItemTypes[type];
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("Work Item With Field Not Found: WorkItemType: {0} - WorkItemTeamProject: {1}.", type, this.WorkItemTeamProject), exception);
            }
        }

        private void SaveWorkItem(WorkItem workItem)
        {
            try
            {
                workItem.Save();
                this.WorkItemId = workItem.Id;
            }
            catch (Exception exception)
            {
                ArrayList list = workItem.Validate();
                if (list.Count > 0)
                {
                    StringBuilder builder = new StringBuilder();
                    for (int i = 0; i < list.Count; i++)
                    {
                        Field field = (Field)list[i];
                        if (i != 0)
                        {
                            builder.Append(CultureInfo.CurrentCulture.TextInfo.ListSeparator);
                        }
                        builder.Append(string.Format("Field Error: {0} : {1}.", new object[] { field.Name, field.Value }));
                    }
                    throw new Exception(string.Format("Field Errors Saving New Work Item: {0}.", new object[] { builder.ToString() }), exception);
                }
                if (!workItem.Project.HasWorkItemWriteRights)
                {
                    throw new Exception("No Permission To Write WI.", exception);
                }
                throw;
            }
        }

        private void SetFieldIfExists(WorkItem workItem, string fieldRefName, string value)
        {
            if (workItem.Type.FieldDefinitions.Contains(fieldRefName))
            {
                Field field = workItem.Fields[fieldRefName];
                if (field.IsEditable && (!field.IsLimitedToAllowedValues || field.AllowedValues.Contains(value)))
                {
                    workItem[fieldRefName] = value;
                }
            }
        }

        private void UpdateUserSpecifiedFields(WorkItem workItem, Dictionary<string, string> propertiesTable)
        {
            bool flag = false;
            if (propertiesTable != null)
            {
                foreach (KeyValuePair<string, string> pair in propertiesTable)
                {
                    try
                    {
                        Field field = workItem.Fields[pair.Key];
                        field.Value = pair.Value;
                        if (TFStringComparer.WorkItemFieldReferenceName.Equals(field.ReferenceName, CoreFieldReferenceNames.Reason))
                        {
                            flag = true;
                        }
                        continue;
                    }
                    catch (Exception exception)
                    {
                        Log.LogWarning(string.Format("Unable To Update WI Field: {0} - {1}.", new object[] { pair.Key, exception.Message }));
                        continue;
                    }
                }
            }
            if (!flag)
            {
                this.SetFieldIfExists(workItem, CoreFieldReferenceNames.Reason, DEFAULT_REASON);
            }
        }

        private void NonFatalError(object sender, ExceptionEventArgs e)
        {
            string errorMessage = null;
            if (e.Exception != null)
            {
                errorMessage = e.Exception.Message;
            }
            if (e.Failure != null)
            {
                if (e.Failure.Severity == SeverityType.Warning)
                {
                    return;
                }
                errorMessage = e.Failure.Message;
            }
            Log.LogError(errorMessage);
            this._numErrors++;
        }

        #endregion

        #region Properties

        public string AssignedTo
        {
            get
            {
                return this._assignedTo;
            }
            set
            {
                this._assignedTo = value;
            }
        }

        public int ChangesetId
        {
            get
            {
                return this._changesetId;
            }
            set
            {
                this._changesetId = value;
            }
        }

        [Required]
        public string Description
        {
            get
            {
                return this._description;
            }
            set
            {
                this._description = value;
            }
        }

        [Required]
        public string WorkItemTeamProject
        {
            get
            {
                return this._workItemTeamProject;
            }
            set
            {
                this._workItemTeamProject = value;
            }
        }

        [Required]
        public string Title
        {
            get
            {
                return this._title;
            }
            set
            {
                this._title = value;
            }
        }

        public string AssociatedWorkItems
        {
            get
            {
                return _associatedWorkItems;
            }
            set
            {
                _associatedWorkItems = value;
            }
        }

        public string WorkItemFieldValues
        {
            get
            {
                return this._properties;
            }
            set
            {
                this._properties = value;
            }
        }

        internal WorkItemStore WorkItemStore
        {
            get
            {
                if (this._workItemStore == null)
                {
                    _workItemStore = (WorkItemStore)this.Tfs.GetService(typeof(WorkItemStore));
                }
                return _workItemStore;
            }
        }

        public string WorkItemType
        {
            get
            {
                return this._workItemTypeName;
            }
            set
            {
                this._workItemTypeName = value;
            }
        }

        [Required]
        public string TeamFoundationServerUrl
        {
            get
            {
                return this._tfsUrl;
            }
            set
            {
                this._tfsUrl = value;
            }
        }

        internal TfsTeamProjectCollection Tfs
        {
            get
            {
                if (this._tfs == null)
                {
                    this._tfs = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(new Uri(this.TeamFoundationServerUrl));
                }
                return this._tfs;
            }
        }

        internal VersionControlServer VersionControlServer
        {
            get
            {
                if (this._versionControlServer == null)
                {
                    this._versionControlServer = (VersionControlServer)this.Tfs.GetService(typeof(VersionControlServer));
                    this._versionControlServer.NonFatalError += new ExceptionEventHandler(this.NonFatalError);
                }
                return this._versionControlServer;
            }
        }

        public string[] AttachmentPaths
        {
            get
            {
                return _attachmentPaths;
            }
            set
            {
                _attachmentPaths = value;
            }
        }

        [Output]
        public int WorkItemId
        {
            get
            {
                return this._workItemId;
            }
            set
            {
                this._workItemId = value;
            }
        }

        #endregion
    }
}
