﻿using System;
using System.Threading;
using System.Diagnostics;

namespace MsdnMag.Exchange.PushNotificationClient
{
    /// <summary>
    /// Class for managing the retry logic around operations.
    /// </summary>
    public static class RetryManager
    {
        private const int DEFAULT_RETRIES_COUNT = 3;
        private const int WAIT_TIME_MILLISECONDS = 1000;

        /// <summary>
        /// Invoke the method with no parameters and a default retry limit.
        /// </summary>
        /// <param name="action"></param>
        public static void InvokeWithRetries(Action action)
        {
            InvokeWithRetries(action, DEFAULT_RETRIES_COUNT);
        }

        /// <summary>
        /// Invoke the method with no parameters and the specified retry limit.
        /// </summary>
        /// <param name="action"></param>
        /// <param name="numberOfRetries"></param>
        public static void InvokeWithRetries(Action action, int numberOfRetries)
        {
            int currentRetriesCount = 0;
            while (currentRetriesCount < numberOfRetries)
            {
                currentRetriesCount++;

                try
                {
                    action.Invoke();

                    return;
                }
                catch (Exception ex)
                {
                    if (currentRetriesCount < numberOfRetries)
                    {
                        Trace.WriteLine(string.Format("InvokeWithRetries: Error: {0}. Current attempt: {1}. Sleeping for {2} milliseconds before retrying...",
                            ex.Message,
                            currentRetriesCount,
                            WAIT_TIME_MILLISECONDS));

                        Thread.Sleep(WAIT_TIME_MILLISECONDS);

                        continue;
                    }
                    else
                    {
                        Trace.WriteLine(string.Format("InvokeWithRetries: Error: {0}. Current attempt: {1}. Retries expired. Rethrowing exception...",
                            ex.Message,
                            currentRetriesCount,
                            WAIT_TIME_MILLISECONDS));

                        throw;
                    }
                }
            }
        }
    }
}
