﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.Build.Utilities;
using Microsoft.Build.Framework;
using System.Collections;
using System.Globalization;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace MsdnMag.Exchange.PushNotificationClient
{
    /// <summary>
    /// Updates the specified fields of a given work item.
    /// </summary>
    public class UpdateWorkItem : Task
    {
        #region Private Fields

        private string _properties;
        private string _workItemTeamProject;
        private string _tfsUrl;
        private int _workItemId;
        private string _history;
        private int _numErrors;
        private TfsTeamProjectCollection _tfs;
        private VersionControlServer _versionControlServer;
        private WorkItemStore _workItemStore;
        private string _associatedWorkItems;
        private string _associatedChangesets;

        #endregion

        #region Methods

        public override bool Execute()
        {
            WorkItem workItem = this.WorkItemStore.GetWorkItem(_workItemId);

            this.UpdateUserSpecifiedFields(workItem, PropertyParser.GetTable(this.WorkItemFieldValues));

            if (!string.IsNullOrWhiteSpace(this.AssociatedWorkItems))
            {
                this.AddLinkToAssociatedWorkItems(workItem, this.AssociatedWorkItems);
            }

            if (!string.IsNullOrWhiteSpace(this.AssociatedChangesets))
            {
                this.AddLinkToAssociatedChangesets(workItem, this.AssociatedChangesets);
            }

            // Update the history:
            if (!string.IsNullOrWhiteSpace(_history))
            {
                workItem.History = _history;
            }

            this.SaveWorkItem(workItem);

            Log.LogMessage("UpdateWorkItem: Workitem Updated: Id: {0}.", new object[] { workItem.Id });

            return true;
        }

        private void AddLinkToAssociatedChangesets(WorkItem workItem, string associatedChangesets)
        {
            string[] associatedChangesetsArray = associatedChangesets.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);

            if (associatedChangesetsArray != null && associatedChangesetsArray.Length > 0)
            {
                foreach (var item in associatedChangesetsArray)
                {
                    try
                    {
                        Changeset changeset = this.VersionControlServer.GetChangeset(Convert.ToInt32(item));
                        Link link = new ExternalLink(this.WorkItemStore.RegisteredLinkTypes["Source Code File"], changeset.ArtifactUri.AbsoluteUri);
                        workItem.Links.Add(link);

                        Log.LogMessage(string.Format("UpdateWorkItem: Added a link to associated changeset: {0}.", new object[] { changeset.ChangesetId }));
                    }
                    catch (Exception ex)
                    {
                        Log.LogWarning(string.Format("Unable to associate changeset: {0}.\r\nError Message: {1}\r\nStackTrace: {2}.", new object[] { item, ex.Message, ex.StackTrace }));
                    }
                }
            }
            else
            {
                Log.LogWarning("UpdateWorkItem: AssociatedWorkItems list is empty.");
                return;
            }
        }

        private void AddLinkToAssociatedWorkItems(WorkItem workItem, string associatedWorkItems)
        {
            string[] associatedWorkItemsArray = associatedWorkItems.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);

            if (associatedWorkItemsArray != null && associatedWorkItemsArray.Length > 0)
            {
                foreach (var item in associatedWorkItemsArray)
                {
                    try
                    {
                        int workItemId = Convert.ToInt32(item);

                        if (!IsWorkItemAlreadyLinked(workItemId, workItem.Links))
                        {
                            RelatedLink link = new RelatedLink(workItemId);
                            workItem.Links.Add(link);

                            Log.LogMessage(string.Format("UpdateWorkItem: Added a link to associated work item: {0}.", new object[] { workItemId }));
                        }
                        else
                        {
                            Log.LogMessage(string.Format("UpdateWorkItem: Skipped linking to associated work item as it is already linked: {0}.", new object[] { workItemId }));
                        }
                    }
                    catch (Exception ex)
                    {
                        Log.LogWarning(string.Format("Unable to associate work item: {0}.\r\nError Message: {1}\r\nStackTrace: {2}.", new object[] { item, ex.Message, ex.StackTrace }));
                    }
                }
            }
            else
            {
                Log.LogWarning("UpdateWorkItem: AssociatedWorkItems list is empty.");
                return;
            }
        }

        internal static bool IsWorkItemAlreadyLinked(int workItemId, LinkCollection links)
        {
            if (workItemId == 0 || links == null || links.Count == 0)
                return false;

            foreach (Link link in links)
            {
                if (link is RelatedLink && ((RelatedLink)link).RelatedWorkItemId == workItemId)
                    return true;
            }
            return false;
        }

        private void SaveWorkItem(WorkItem workItem)
        {
            try
            {
                workItem.Save();
                this.WorkItemId = workItem.Id;
            }
            catch (Exception exception)
            {
                ArrayList list = workItem.Validate();
                if (list.Count > 0)
                {
                    StringBuilder builder = new StringBuilder();
                    for (int i = 0; i < list.Count; i++)
                    {
                        Field field = (Field)list[i];
                        if (i != 0)
                        {
                            builder.Append(CultureInfo.CurrentCulture.TextInfo.ListSeparator);
                        }
                        builder.Append(string.Format("Field Error: {0} : {1}.", new object[] { field.Name, field.Value }));
                    }
                    throw new Exception(string.Format("Field Errors Saving New Work Item: {0}.", new object[] { builder.ToString() }), exception);
                }
                if (!workItem.Project.HasWorkItemWriteRights)
                {
                    throw new Exception("No Permission To Write WI.", exception);
                }
                throw;
            }
        }

        private void SetFieldIfExists(WorkItem workItem, string fieldRefName, string value)
        {
            if (workItem.Type.FieldDefinitions.Contains(fieldRefName))
            {
                Field field = workItem.Fields[fieldRefName];
                if (field.IsEditable && (!field.IsLimitedToAllowedValues || field.AllowedValues.Contains(value)))
                {
                    workItem[fieldRefName] = value;
                }
            }
        }

        private void UpdateUserSpecifiedFields(WorkItem workItem, Dictionary<string, string> propertiesTable)
        {
            if (propertiesTable != null)
            {
                foreach (KeyValuePair<string, string> pair in propertiesTable)
                {
                    try
                    {
                        Field field = workItem.Fields[pair.Key];
                        field.Value = pair.Value;
                    }
                    catch (Exception exception)
                    {
                        Log.LogWarning(string.Format("Unable To Update WI Field: {0} - {1}.", new object[] { pair.Key, exception.Message }));
                        continue;
                    }
                }
            }
        }

        private void NonFatalError(object sender, ExceptionEventArgs e)
        {
            string errorMessage = null;
            if (e.Exception != null)
            {
                errorMessage = e.Exception.Message;
            }
            if (e.Failure != null)
            {
                if (e.Failure.Severity == SeverityType.Warning)
                {
                    return;
                }
                errorMessage = e.Failure.Message;
            }
            Log.LogError(errorMessage);
            this._numErrors++;
        }

        #endregion

        #region Properties

        [Required]
        public string WorkItemTeamProject
        {
            get
            {
                return this._workItemTeamProject;
            }
            set
            {
                this._workItemTeamProject = value;
            }
        }

        public string AssociatedWorkItems
        {
            get
            {
                return _associatedWorkItems;
            }
            set
            {
                _associatedWorkItems = value;
            }
        }

        public string AssociatedChangesets
        {
            get
            {
                return _associatedChangesets;
            }
            set
            {
                _associatedChangesets = value;
            }
        }

        public string WorkItemFieldValues
        {
            get
            {
                return this._properties;
            }
            set
            {
                this._properties = value;
            }
        }

        public string History
        {
            get
            {
                return _history;
            }
            set
            {
                _history = value;
            }
        }

        internal WorkItemStore WorkItemStore
        {
            get
            {
                if (this._workItemStore == null)
                {
                    _workItemStore = (WorkItemStore)this.Tfs.GetService(typeof(WorkItemStore));
                }
                return _workItemStore;
            }
        }

        [Required]
        public string TeamFoundationServerUrl
        {
            get
            {
                return this._tfsUrl;
            }
            set
            {
                this._tfsUrl = value;
            }
        }

        protected TfsTeamProjectCollection Tfs
        {
            get
            {
                if (this._tfs == null)
                {
                    this._tfs = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(new Uri(this.TeamFoundationServerUrl));
                }
                return this._tfs;
            }
        }

        internal VersionControlServer VersionControlServer
        {
            get
            {
                if (this._versionControlServer == null)
                {
                    this._versionControlServer = (VersionControlServer)this.Tfs.GetService(typeof(VersionControlServer));
                    this._versionControlServer.NonFatalError += new ExceptionEventHandler(this.NonFatalError);
                }
                return this._versionControlServer;
            }
        }

        [Required]
        public int WorkItemId
        {
            get
            {
                return this._workItemId;
            }
            set
            {
                this._workItemId = value;
            }
        }

        #endregion
    }
}
