﻿using System;
using MsdnMag.Exchange.PushNotificationSubscriber.Core.EWSReference;
using System.Net;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Net.Security;
using System.Xml.Serialization;
using Microsoft.Exchange.WebServices.Data;

namespace MsdnMag.Exchange.PushNotificationSubscriber
{
    public class PushNotificationSubscriber
    {
        #region Constants

        private const string MAILBOX_EVENTLOG_TEMPLATE = "{0}MailboxEventLog-{1}.txt";
        private const string TIMESTAMP_TEMPLATE = "{0}_{1}_{2}-{3}_{4}_{5}_{6}";
        private const string CREATE_XML_MESSAGE_TEMPLATE = "Writing request and response XML to disk...";
        private const string REQUEST_XML_FILENAME_TEMPLATE = "{0}{1}-{2}.xml";
        private const string RESPONSE_XML_FILENAME_TEMPLATE = "{0}{1}-{2}.xml";
        private const string SUBSCRIBED_MESSAGE_TEMPLATE = "Subscribed for Push notifications: {0}";
        private const string UNSUBSCRIBED_MESSAGE_TEMPLATE = "UnSubscribed from Push notifications. {0}";
        private const string SUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE = "\r\nSubscribed for Push notifications.\r\nSubscription Id: {0}";
        private const string UNSUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE = "UnSubscribed from Push notifications.\r\nSubscription Id: {0}";
        private const string SUBSCRIBED_FAILURE_MESSAGE_TEMPLATE = "\r\nSubscription for Push notifications was not successful.";
        private const string UNSUBSCRIBED_FAILURE_MESSAGE_TEMPLATE = "\r\nUnSubscription from Push notifications was not successful.";

        #endregion

        #region Constructors

        public PushNotificationSubscriber()
        {
        }

        #endregion

        #region Public Properties

        public string EwsServiceUrl { get; set; }
        public string PushClientWebServiceUrl { get; set; }
        public string BaseLogFolder { get; set; }
        public int PushNotificationsFrequencyInMinutes { get; set; }
        public string SubscriptionId { get; private set; }
        public string UserName { get; set; }
        public string Password { get; set; }
        public string Domain { get; set; }
        public bool UseAutodiscoveryForSubscribing { get; set; }
        public string AutoDiscoveryEmail { get; set; }
        public bool UseCurrentlyLoggedInUserCredentials { get; set; }

        #endregion

        #region Public Methods

        /// <summary>
        /// Creates a push notification subscription.
        /// </summary>
        public void SubscribeForPushNotifications()
        {
            SubscribeUnsubscribe(false, null);
        }

        /// <summary>
        /// Unsubscribes an existing push notification subscription.
        /// </summary>
        /// <param name="subscriptionId">The Subscription Id of the push notification subscription to be unsubscribed.</param>
        public void UnSubscribeFromPushNotifications(string subscriptionId)
        {
            SubscribeUnsubscribe(true, subscriptionId);
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// A proxy method that prepares the binding information and either subscribes to a push notification or unsubscribes from one.
        /// </summary>
        /// <param name="unSubscribe">A boolean flag that indicates whether this is a subscription or an unsubscription.</param>
        /// <param name="subscriptionId">The Subscription Id of an existing subscription when the unSubscribe flag is set to true.
        /// Set to null when unSubscribe is set to false.
        /// </param>
        private void SubscribeUnsubscribe(bool unSubscribe, string subscriptionId)
        {
            System.Net.ServicePointManager.ServerCertificateValidationCallback =
                delegate(Object obj, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors)
                {
                    // Replace this line with code to validate server certificate.
                    return true;
                };

            // Create the bindings and set the credentials.
            ExchangeServiceBinding esb = new ExchangeServiceBinding();
            esb.RequestServerVersionValue = new RequestServerVersion();
            esb.RequestServerVersionValue.Version = ExchangeVersionType.Exchange2010;
            esb.Url = EwsServiceUrl;
            if (UseCurrentlyLoggedInUserCredentials)
            {
                esb.Credentials = CredentialCache.DefaultNetworkCredentials;
            }
            else
            {
                esb.Credentials = new NetworkCredential(UserName, Password, Domain);
            }

            if (UseAutodiscoveryForSubscribing)
            {
                ExchangeService service = new ExchangeService();
                service.Credentials = new WebCredentials(esb.Credentials);
                service.AutodiscoverUrl(AutoDiscoveryEmail);
                esb.Url = service.Url.ToString();
            }

            if (unSubscribe)
            {
                Unsubscribe(esb, subscriptionId);
            }
            else
            {
                Subscribe(esb);
            }
        }

        /// <summary>
        /// Creates a subscription for the ExchangeServiceBinding instance. 
        /// </summary>
        /// <param name="esb">An ExchangeServiceBinding instance containing the binding information.</param>
        private void Subscribe(ExchangeServiceBinding esb)
        {
            // Create a new subscription.
            SubscribeType subscribeRequest = new SubscribeType();
            PushSubscriptionRequestType pushSubscription = new PushSubscriptionRequestType();

            // Subscribe to events in the inbox folder.
            BaseFolderIdType[] folders = new BaseFolderIdType[1];
            DistinguishedFolderIdType folderId = new DistinguishedFolderIdType();
            folderId.Id = DistinguishedFolderIdNameType.inbox;
            folders[0] = folderId;
            pushSubscription.FolderIds = folders;

            // Subscribe to new mail event only.
            NotificationEventTypeType[] eventTypes = new NotificationEventTypeType[1];
            eventTypes[0] = NotificationEventTypeType.NewMailEvent;

            pushSubscription.EventTypes = eventTypes;

            // Receive push notifications every n minutes (typically 1 minute).
            pushSubscription.StatusFrequency = PushNotificationsFrequencyInMinutes;

            // Identify the location of the client Web service.
            pushSubscription.URL = PushClientWebServiceUrl;

            // Form the subscribe request.
            subscribeRequest.Item = pushSubscription;

            // Send the subscribe request and get the response.
            SubscribeResponseType subscribeResponse = esb.Subscribe(subscribeRequest);

            string timestamp = GetTimestamp();

            // Check the result.
            if (subscribeResponse.ResponseMessages.Items.Length > 0 &&
               subscribeResponse.ResponseMessages.Items[0].ResponseClass == ResponseClassType.Success)
            {
                SubscribeResponseMessageType subscribeResponseMessage =
                    subscribeResponse.ResponseMessages.Items[0] as SubscribeResponseMessageType;

                SubscriptionId = subscribeResponseMessage.SubscriptionId;

                if (!string.IsNullOrEmpty(BaseLogFolder))
                {
                    using (StreamWriter sw = new StreamWriter(string.Format(MAILBOX_EVENTLOG_TEMPLATE, BaseLogFolder, timestamp)))
                    {
                        sw.WriteLine(SUBSCRIBED_MESSAGE_TEMPLATE, SubscriptionId);
                    }
                }

                Console.WriteLine(SUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE, SubscriptionId);
            }
            else
            {
                Console.WriteLine(SUBSCRIBED_FAILURE_MESSAGE_TEMPLATE);
            }

            CreateXmlMessageTextFile(subscribeRequest, subscribeResponse, BaseLogFolder, timestamp);
        }

        /// <summary>
        /// Unsubscribes an existing subscription.
        /// </summary>
        /// <param name="esb">An ExchangeServiceBinding instance containing the binding information.</param>
        /// <param name="subscriptionId">The Subscription Id for the subscription to be unsubscribed.</param>
        private void Unsubscribe(ExchangeServiceBinding esb, string subscriptionId)
        {
            // Create a new Unsubscription request.
            UnsubscribeType unSubscribeRequest = new UnsubscribeType();

            // Form the subscribe request.
            unSubscribeRequest.SubscriptionId = subscriptionId;

            // Send the subscribe request and get the response.
            UnsubscribeResponseType unSubscribeResponse = esb.Unsubscribe(unSubscribeRequest);

            string timestamp = GetTimestamp();

            // Check the result.
            if (unSubscribeResponse.ResponseMessages.Items.Length > 0 &&
               unSubscribeResponse.ResponseMessages.Items[0].ResponseClass == ResponseClassType.Success)
            {
                SubscribeResponseMessageType subscribeResponseMessage =
                    unSubscribeResponse.ResponseMessages.Items[0] as SubscribeResponseMessageType;

                if (!string.IsNullOrEmpty(BaseLogFolder))
                {
                    using (StreamWriter sw = new StreamWriter(string.Format(MAILBOX_EVENTLOG_TEMPLATE, BaseLogFolder, timestamp)))
                    {
                        sw.WriteLine(UNSUBSCRIBED_MESSAGE_TEMPLATE, subscribeResponseMessage.SubscriptionId);
                    }
                }

                Console.WriteLine(UNSUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE, subscribeResponseMessage.SubscriptionId);
            }
            else
            {
                Console.WriteLine(UNSUBSCRIBED_FAILURE_MESSAGE_TEMPLATE);
            }

            CreateXmlMessageTextFile(unSubscribeRequest, unSubscribeResponse, BaseLogFolder, timestamp);
        }

        #endregion

        #region Static Helper Methods

        /// <summary>
        /// Persists the request (BaseRequestType) and response (BaseResponseMessageType) messages to XML files.
        /// </summary>
        /// <param name="request">An instance of BaseRequestType.</param>
        /// <param name="response">An instance of BaseResponseMessageType.</param>
        /// <param name="baseLogFolder">The folder where the XML files are to be written.</param>
        /// <param name="timestamp">A formatted timestamp value with which the XML files will be suffixed.</param>
        private static void CreateXmlMessageTextFile(BaseRequestType request, BaseResponseMessageType response, string baseLogFolder, string timestamp)
        {
            if (string.IsNullOrEmpty(baseLogFolder))
            {
                return;
            }

            Console.WriteLine(CREATE_XML_MESSAGE_TEMPLATE);

            // Create a text file with the request XML.
            using (StreamWriter myReqWriter = new StreamWriter(string.Format(REQUEST_XML_FILENAME_TEMPLATE, baseLogFolder, request.GetType().Name, timestamp)))
            {
                XmlSerializer mySerializer = new XmlSerializer(request.GetType());
                mySerializer.Serialize(myReqWriter, request);
            }

            // Create a text file with the response XML.
            using (StreamWriter myRespWriter = new StreamWriter(string.Format(RESPONSE_XML_FILENAME_TEMPLATE, baseLogFolder, response.GetType().Name, timestamp)))
            {
                XmlSerializer mySerializer = new XmlSerializer(response.GetType());
                mySerializer.Serialize(myRespWriter, response);
            }
        }

        /// <summary>
        /// Generates a timestamp string to be used as the prefix or suffix of filenames that need to be uniquely distinguished.
        /// </summary>
        /// <returns>A formatted string containing the current datetime.</returns>
        public static string GetTimestamp()
        {
            DateTime currentTime = DateTime.Now;
            return string.Format(TIMESTAMP_TEMPLATE,
                currentTime.Year,
                currentTime.Month,
                currentTime.Day,
                currentTime.Hour,
                currentTime.Minute,
                currentTime.Second,
                currentTime.Millisecond);
        }

        #endregion
    }
}
